from typing import Dict, List, Union

import numpy as np
from bilby.core.prior import PriorDict, Uniform
from scipy.stats import qmc
from scipy.stats.qmc import LatinHypercube

DEFAULT_SF_PARAMETERS = dict(
    aSF=0.01,
    bSF=2.77,
    cSF=2.9,
    dSF=4.7,
    mu_z=-0.23,
    sigma_0=0.39,
)
DEFAULT_DICT = {
    k: DEFAULT_SF_PARAMETERS[k] for k in ["aSF", "dSF", "mu_z", "sigma_0"]
}

STAR_FORMATION_RANGES = dict(
    mu_z=[-0.5, -0.001],  # Jeff's alpha
    sigma_0=[0.1, 0.6],  # Jeff's sigma
    aSF=[0.005, 0.015],
    dSF=[4.2, 5.2],
)
# STAR_FORMATION_RANGES = dict(
#     mu_z=[-2, -0.001],  # Jeff's alpha
#     sigma_0=[0.1, 2],  # Jeff's sigma
#     aSF=[0.005, 0.2],
#     dSF=[3, 5.2],
# )

LATEX_LABELS = dict(
    mu_z=r"$\mu_z$",
    sigma_0=r"$\sigma_0$",
    aSF=r"$\rm{SF}[a]$",
    dSF=r"$\rm{SF}[d]$",
)


class MyPriorDict(PriorDict):
    @property
    def bounds(self):
        return np.array(
            [[self[p].minimum, self[p].maximum] for p in self.keys()]
        )

    @property
    def n_params(self):
        return len(self.keys())

    def sample_val(self):
        return np.array([self[p].sample() for p in self.keys()]).reshape(1, -1)


def get_star_formation_prior(parameters=None) -> MyPriorDict:
    if parameters is None:
        parameters = list(STAR_FORMATION_RANGES.keys())
    pri = dict()
    for p in parameters:
        pri[p] = Uniform(
            *STAR_FORMATION_RANGES[p], name=p, latex_label=LATEX_LABELS[p]
        )
    return MyPriorDict(pri)


def draw_star_formation_samples(
    n: int = 1000,
    parameters: List[str] = None,
    as_list=False,
    custom_ranges: Dict = None,
    grid: bool = False,
) -> Union[Dict[str, np.ndarray], List[Dict]]:
    """Draw samples from the star formation parameters.
    Returns a dictionary of arrays, or a list of dictionaries if as_list is True.
    """
    if parameters is None:
        parameters = list(STAR_FORMATION_RANGES.keys())
    if isinstance(parameters, str):
        parameters = [parameters]
    assert all(
        [p in STAR_FORMATION_RANGES for p in parameters]
    ), f"Invalid parameters provided ({parameters} not in {STAR_FORMATION_RANGES.keys()}))"
    num_dim = len(parameters)

    ranges = STAR_FORMATION_RANGES.copy()
    if custom_ranges is not None:
        ranges.update(custom_ranges)
    parameter_ranges = np.array([ranges[p] for p in parameters])
    lower_bound = parameter_ranges[:, 0]
    upper_bound = parameter_ranges[:, 1]

    if grid:
        n_per_dim = int(np.ceil(n ** (1 / num_dim)))
        grid = np.meshgrid(
            *[np.linspace(*r, n_per_dim) for r in parameter_ranges]
        )
        samples = np.vstack([g.ravel() for g in grid]).T

    else:
        sampler = LatinHypercube(d=num_dim)
        unscaled_samples = sampler.random(n)
        samples = qmc.scale(
            unscaled_samples, l_bounds=lower_bound, u_bounds=upper_bound
        )

    dict_of_params = {p: samples[:, i] for i, p in enumerate(parameters)}
    if as_list:
        return [
            dict(zip(dict_of_params, t)) for t in zip(*dict_of_params.values())
        ]
    return dict_of_params


def get_latex_labels(parameters) -> Union[str, List[str]]:
    if isinstance(parameters, str):
        return LATEX_LABELS[parameters]
    else:
        return [LATEX_LABELS[p] for p in parameters]



import glob
import os

import numpy as np
import pytest
from scipy.stats import multivariate_normal

from lnl_computer.cosmic_integration.mcz_grid import McZGrid
from lnl_computer.mock_data import MockData, generate_mock_data

HERE = os.path.dirname(__file__)
TEST_DIR = os.path.join(HERE, "test_data")


@pytest.fixture
def mock_data() -> MockData:
    np.random.seed(42)
    return generate_mock_data(outdir=TEST_DIR, duration=1)


@pytest.fixture
def tmp_path() -> str:
    """Temporary directory."""
    pth = os.path.join(HERE, "out_tmp")
    os.makedirs(pth, exist_ok=True)
    return pth


@pytest.fixture
def monkeypatched_mcz_grid(mock_data: MockData, monkeypatch) -> McZGrid:
    model = mock_data.mcz_grid
    norm2d = multivariate_normal(
        [0.2, 20], np.array([[1.0, 0.0], [0.0, 20.0]])
    )
    z_bins, mc_bins = model.redshift_bins, model.chirp_mass_bins
    Z, MC = np.meshgrid(z_bins, mc_bins)
    model.rate_matrix = norm2d.pdf(np.stack([Z, MC], axis=-1))
    return model



import os

import numpy as np
import pytest

from lnl_computer.cosmic_integration.mcz_grid import McZGrid
from lnl_computer.mock_data import MockData
from lnl_computer.observation.lvk_observation import LVKObservation
from lnl_computer.observation.mock_observation import MockObservation

PLOT = False


def test_lnl(mock_data: MockData):
    lnl, unc = mock_data.mcz_grid.get_lnl(mock_data.observations)
    assert lnl > -np.inf
    assert unc != np.nan

    # NO-BOOTSTRAPS LNL
    lnl, unc = McZGrid.lnl(
        mcz_obs=mock_data.observations,
        compas_h5_path=mock_data.compas_filename,
        sf_sample=dict(aSF=0.01, dSF=4.70, mu_z=-0.23),
        n_bootstraps=0,
    )
    assert lnl > -np.inf
    assert np.isnan(unc)
    expected_lnl = 10647.1
    np.testing.assert_allclose(
        lnl,
        expected_lnl,
        atol=200,
        err_msg=f"lnl={lnl:.1f} not close to {expected_lnl}",
    )


@pytest.mark.skip(reason="Takes too long")
def test_lnl_nan(mock_data: MockData, tmp_path: str):
    # ensure not getting a nan!
    lnl, unc = McZGrid.lnl(
        mcz_obs=mock_data.observations,
        compas_h5_path=mock_data.compas_filename,
        sf_sample=dict(aSF=0.01, dSF=4.70, mu_z=-0.01, sigma_0=0.0),
        n_bootstraps=0,
        save_plots=True,
        outdir=f"{tmp_path}/nan_lnl",
    )
    assert not np.isnan(lnl)
    assert np.isnan(unc)


def test_duration(mock_data: MockData):
    # ensure duration is used
    mock_d1 = MockObservation.from_mcz_grid(mock_data.mcz_grid, duration=1)
    mock_d2 = MockObservation.from_mcz_grid(mock_data.mcz_grid, duration=2)
    assert mock_d1.n_events < mock_d2.n_events


def test_ogc4_lnl(monkeypatched_mcz_grid: McZGrid, tmp_path: str):
    obs = LVKObservation.from_ogc4_data()
    (
        lnl,
        _,
    ) = monkeypatched_mcz_grid.get_lnl(mcz_obs=obs)
    assert lnl > -np.inf



import traceback

from .observation import Observation


def load_observation(obs: str, **kwargs) -> Observation:

    assert isinstance(obs, str), f"Expected str, got {type(obs)}"

    try:
        if obs.upper() == "LVK":
            from .lvk_observation import LVKObservation

            return LVKObservation.from_ogc4_data(**kwargs)
        elif obs.endswith(".npz"):
            return Observation.load(fname=obs)
        else:
            raise ValueError(f"Unknown observation: {obs}")
    except Exception as e:
        raise ValueError(
            f"Error loading observation [{obs}]: {e} {traceback.format_exc()}"
        )



from typing import Callable, Tuple, Union

import numpy as np

from .observation.observation import Observation


def ln_poisson_likelihood(
    n_obs: float, n_model: float, ignore_factorial=True
) -> float:
    """
    Computes LnL(N_obs | N_model) = N_obs * ln(N_model) - N_model - ln(N_obs!)

    :param n_obs: number of observed events
    :param n_model: number of events predicted by the model
    :param ignore_factorial: ignore the factorial term in the likelihood

    # TODO: Why are we ignoring the factorial term?? It was in Ilya's notes from 2023, but unsure why...

    :return: the log likelihood
    """
    if n_model <= 0:
        return -np.inf
    lnl = n_obs * np.log(n_model) - n_model
    if ignore_factorial is False:
        lnl += -np.log(np.math.factorial(n_obs))
    return lnl


def ln_mcz_grid_likelihood(
    obs_weights: np.ndarray, model_prob_grid: np.ndarray
) -> float:
    """
    Computes LnL(mc, z | model)
        =  sum_n  ln sum_i  p(mc_i, z_i | model) * wi,n
        (for N_obs events, and i {mc,z} bins)


    for even_idx in range(n_events):
        p_event = 0
        for mc_idx in range(n_mc):
            for z_idx in range(n_z):
                w = obs_weights[even_idx, mc_idx, z_idx]
                p_event += w * model_prob_grid[mc_idx, z_idx]
        lnl += np.log(p_event)

    """
    p_events = np.einsum("nij,ij->n", obs_weights, model_prob_grid)
    return np.nansum(np.log(p_events))


def ln_likelihood(
    obs: Observation,
    model: "McZGrid",
    detailed=False,
) -> Union[float, Tuple[float, float, float, float]]:
    """
    Compute the log likelihood of the model given the observation

    :param obs: the observation
    :param model: the McZ-grid model
    :param detailed: return detailed likelihood components

    :return: the log likelihood
    if detailed:
        return [lnl, poisson_lnl, mcz_lnl, model_n_detections]

    """

    # unpack the model into the grid and the number of detections
    model_grid = model.get_prob_grid(obs.duration)
    model_n_detections = model.n_detections(obs.duration)

    # compute the likelihood
    poisson_lnl = ln_poisson_likelihood(obs.n_events, model_n_detections)
    mcz_lnl = ln_mcz_grid_likelihood(obs.weights, model_grid)
    lnl = poisson_lnl + mcz_lnl
    if detailed:
        return np.array([lnl, poisson_lnl, mcz_lnl, model_n_detections])
    return lnl



import os

import pandas as pd

from lnl_computer.cosmic_integration.star_formation_paramters import (
    draw_star_formation_samples,
    get_star_formation_prior,
)

PLOT = True


def test_prior():
    p = get_star_formation_prior(parameters=["aSF", "dSF"])
    assert p.bounds.shape == (2, 2)
    p = get_star_formation_prior()
    assert p.bounds.shape == (p.n_params, 2)
    assert p.sample_val().shape == (1, p.n_params)


def test_sampler(tmp_path):
    parameters = ["aSF", "dSF"]
    samples_1000 = draw_star_formation_samples(1000, parameters)
    assert samples_1000["aSF"].shape == (1000,)
    assert samples_1000["dSF"].shape == (1000,)

    samples_30 = pd.DataFrame(
        draw_star_formation_samples(30, parameters, as_list=True)
    )

    samples_grid = pd.DataFrame(
        draw_star_formation_samples(50, parameters, grid=True)
    )

    samples_all = draw_star_formation_samples(1000)

    if PLOT:
        import matplotlib.pyplot as plt

        plt.plot(
            samples_1000["aSF"],
            samples_1000["dSF"],
            ".",
            color="tab:red",
            zorder=0,
        )
        plt.plot(
            samples_30["aSF"],
            samples_30["dSF"],
            ".",
            color="tab:blue",
            zorder=1,
        )
        plt.plot(
            samples_grid["aSF"],
            samples_grid["dSF"],
            "s",
            color="tab:green",
            zorder=1,
            alpha=0.5,
        )
        plt.grid()
        plt.savefig(os.path.join(tmp_path, "test_sf_samples.png"))



"""Mocking utilities for testing (mocks COMPAS populations, and MCZ obserations)."""
import json
import os
from typing import Dict

from compas_python_utils.cosmic_integration.binned_cosmic_integrator.bbh_population import (
    generate_mock_bbh_population_file,
)

from lnl_computer.cosmic_integration.mcz_grid import McZGrid
from lnl_computer.observation import Observation
from lnl_computer.observation.mock_observation import MockObservation


def generate_mock_data(
    outdir: str, duration: float, sf_params: Dict[str, float] = None
):
    """Generate mock datasets for testing."""
    return MockData.generate_mock_datasets(
        outdir=outdir, duration=duration, sf_params=sf_params
    )


def load_mock_data(outdir: str):
    """Load mock datasets for testing."""
    return MockData(outdir)


class MockData(object):
    """Mocking utilities."""

    def __init__(
        self, outdir: str, duration: float, sf_params: Dict[str, float] = None
    ):
        os.makedirs(outdir, exist_ok=True)
        self.outdir = outdir
        self.duration = duration
        self.sf_params = sf_params or {}

    @property
    def compas_filename(self):
        return os.path.join(self.outdir, "mock_COMPAS_output.h5")

    @property
    def observations_filename(self):
        return os.path.join(self.outdir, "mock_MZC_obs.npz")

    @property
    def mcz_grid_filename(self):
        return os.path.join(self.outdir, "mock_MZC_output.h5")

    @classmethod
    def generate_mock_datasets(
        cls,
        outdir: str,
        duration: float,
        sf_params: Dict[str, float] = None,
    ):
        return cls.load(outdir, duration, sf_params)

    @classmethod
    def load(
        cls,
        outdir: str,
        duration: float,
        sf_params: Dict[str, float] = None,
    ):
        self = cls(outdir, duration, sf_params)
        if not os.path.exists(self.compas_filename):
            generate_mock_bbh_population_file(
                filename=self.compas_filename, frac_bbh=0.05
            )

        if not os.path.exists(self.mcz_grid_filename):
            McZGrid.generate_n_save(
                self.compas_filename,
                sf_sample=sf_params,
                fname=self.mcz_grid_filename,
            )

        if not os.path.exists(self.observations_filename):
            obs = MockObservation.from_mcz_grid(
                self.mcz_grid, duration=duration
            )
            obs.save(self.observations_filename)
        return self

    @property
    def mcz_grid(self) -> McZGrid:
        if not hasattr(self, "_mcz_grid"):
            self._mcz_grid = McZGrid.from_h5(self.mcz_grid_filename)
        return self._mcz_grid

    @property
    def observations(self) -> Observation:
        return Observation.load(fname=self.observations_filename)

    @property
    def truth(self) -> Dict:
        return _get_true_params(self)


def _get_true_params(mock_data: MockData):
    fn = f"{mock_data.outdir}/truth.json"
    if os.path.exists(fn):
        with open(fn, "r") as f:
            return json.load(f)
    else:
        grid = mock_data.mcz_grid
        cosmo_params = grid.cosmological_parameters
        true_params = dict(
            aSF=cosmo_params["aSF"],
            dSF=cosmo_params["dSF"],
            sigma_0=cosmo_params["sigma_0"],
            mu_z=cosmo_params["mu_z"],
        )
        lnl = (
            grid.lnl(
                mcz_obs=mock_data.observations,
                duration=mock_data.duration,
                compas_h5_path=mock_data.compas_filename,
                sf_sample=true_params.copy(),
                n_bootstraps=0,
            )[0]
            * -1
        )
        true_params["lnl"] = lnl
        with open(fn, "w") as f:
            json.dump(true_params, f)
        return true_params



import json

import numpy as np

from lnl_computer.cli.main import make_mock_obs
from lnl_computer.mock_data import (
    McZGrid,
    MockData,
    MockObservation,
    generate_mock_data,
)


def test_reproducible_dataset(tmpdir):
    mock_datapaths = [f"{tmpdir}/test_data{i}" for i in range(2)]
    kwgs = dict(duration=1, sf_params=dict(aSF=0.01, dSF=4.70, mu_z=-0.23))
    data = []
    for path in mock_datapaths:
        np.random.seed(42)
        data.append(generate_mock_data(outdir=path, **kwgs))

    # ensure the McZ grid is the same
    grid0 = McZGrid.from_h5(data[0].mcz_grid_filename)
    grid1 = McZGrid.from_h5(data[1].mcz_grid_filename)
    assert np.allclose(grid0.rate_matrix, grid1.rate_matrix)

    # ensure the observations are the same
    obs0 = MockObservation.load(data[0].observations_filename)
    obs1 = MockObservation.load(data[1].observations_filename)
    assert np.allclose(obs0.weights, obs1.weights)

    # ensure the reference_param is the same
    truth0 = data[0].truth
    truth1 = data[1].truth
    assert truth0 == truth1


def test_mock_obs_lnl(tmpdir, mock_data: MockData):
    cosmo_params = dict(aSF=0.01, dSF=4.70, mu_z=-0.23)
    obs_fname = f"{tmpdir}/mock_obs.npz"
    make_mock_obs(
        compas_h5_path=mock_data.compas_filename,
        sf_sample=cosmo_params,
        duration=1,
        fname=obs_fname,
    )
    truths_fn = f"{tmpdir}/reference_param.json"
    # load the reference_param
    with open(truths_fn, "r") as f:
        truth = json.load(f)
    assert isinstance(
        truth["lnl"], float
    ), f"truth['lnl'] not a float: {truth['lnl']}"

    # load the observation
    obs = MockObservation.load(obs_fname)
    # assert dict compare
    assert obs.cosmological_parameters["aSF"] == cosmo_params["aSF"]
    assert obs.cosmological_parameters["dSF"] == cosmo_params["dSF"]



import logging

import colorlog
import tqdm


class TqdmLoggingHandler(logging.StreamHandler):
    def __init__(self):
        logging.StreamHandler.__init__(self)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg)
            self.flush()
        except Exception:
            self.handleError(record)


def setup_logger(name: str):
    logger = colorlog.getLogger(name)
    logger.setLevel(logging.INFO)
    handler = TqdmLoggingHandler()
    handler.setFormatter(
        colorlog.ColoredFormatter(
            "%(log_color)s%(name)s | %(asctime)s | %(levelname)s | %(message)s",
            datefmt="%d/%m/%y %H:%M:%S",
            log_colors={
                "DEBUG": "cyan",
                "INFO": "white",
                "SUCCESS:": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red,bg_white",
            },
        )
    )

    logger.addHandler(handler)
    return logger


logger = setup_logger("lnl_computer")



import os
import uuid
from argparse import ArgumentParser
from functools import partial
from glob import glob
from multiprocessing import cpu_count
from typing import Dict, List, Tuple, Union

import click
import corner
import numpy as np
import pandas as pd
from tqdm.contrib.concurrent import process_map

from ..cosmic_integration.mcz_grid import McZGrid
from ..cosmic_integration.star_formation_paramters import (
    draw_star_formation_samples,
)
from ..logger import logger
from ..observation import Observation, load_observation
from ..observation.mock_observation import MockObservation


def make_sf_table(
    parameters: List[str] = None,
    n: int = 50,
    grid_parameterspace=False,
    fname: str = "parameter_table.csv",
) -> pd.DataFrame:
    """Parses the table of parameters to generate mcz-grids for.

    :param parameters: list of parameters to generate mcz-grids for
    :param parameter_table: path to pandas dataframe csv containing cosmic integration parameters (or dataframe itself)
    :param n: number of samples to generate
    :param custom_ranges: custom ranges for parameters
    :param grid_parameterspace: whether to grid the parameter space
    :param outdir: output directory for mcz-grids
    :return: pandas dataframe containing cosmic integration parameters
    """
    logger.info(f"Generating {n} samples of {parameters}")
    parameters = parameters or ["aSF", "dSF", "mu_z", "sigma_0"]
    parameter_table = pd.DataFrame(
        draw_star_formation_samples(
            n,
            parameters=parameters,
            as_list=True,
            custom_ranges=None,
            grid=grid_parameterspace,
        )
    )
    # TODO: DOUBLE CHECK WITH JEFF -- mu_z, sigma_z, sigma_0???

    parameter_table.to_csv(fname, index=False)
    fig = corner.corner(
        parameter_table.values,
        labels=parameters,
        truths=None,
        plot_datapoints=True,
        plot_density=False,
        plot_contours=False,
        no_fill_contours=False,
        fill_contours=False,
    )
    fig.savefig(fname.replace(".csv", ".png"))
    logger.info(f"Parameter table saved to {fname}")


def make_mock_obs(
    compas_h5_path: str,
    sf_sample: Union[Dict, str],
    duration: float,
    fname: str = "observation.npz",
    save_plots: bool = False,
) -> "MockObservation":
    """Generate a detection matrix for a given set of star formation parameters
    :param compas_h5_path:
    :param sf_sample: Dict of star formation parameters, or a string like "k1:v1 k2:v2"->{"k1":float(v1)", "k2":float(v2)}
    :param duration: duration of the observation (years)
    :param fname: mcgrid-fname
    """
    assert fname.endswith(".npz"), "fname must end with .npz"
    if isinstance(sf_sample, str):
        sf_sample = dict([s.split(":") for s in sf_sample.split(" ")])
        sf_sample = {k: float(v) for k, v in sf_sample.items()}

    mcz_grid = McZGrid.generate_n_save(
        compas_h5_path=compas_h5_path,
        sf_sample=sf_sample,
        save_plots=save_plots,
    )
    obs = MockObservation.from_mcz_grid(mcz_grid, duration=duration)
    obs.save(fname)

    # save reference_param-json
    lnl = mcz_grid.get_lnl(mcz_obs=obs)
    truth_fname = os.path.dirname(fname) + "/reference_param.json"
    truth_data = dict(lnl=lnl[0], **sf_sample)
    logger.info(f"MockObs reference_param: {truth_data}")
    _write_json(data=truth_data, fname=truth_fname)
    logger.info(
        f"Mock observation saved to {fname} and truths to {truth_fname}"
    )


def _write_json(data, fname):
    import json

    with open(fname, "w") as f:
        json.dump(data, f)


def batch_lnl_generation(
    mcz_obs_path: str,
    compas_h5_path: str,
    parameter_table: Union[pd.DataFrame, str],
    n_bootstraps: int = 100,
    save_images: bool = True,
    outdir: str = "out_mcz_grids",
) -> None:
    """
    Generate a set of COMPAS Mc-Z detection rate matrices
    :param compas_h5_path: path to COMPAS h5 file
    :param parameter_table: path to pandas dataframe containing cosmic integration parameters
    :param n_bootstraps: number of bootstraps to generate for each parameter set
    :param save_images: save images of the generated mcz-grids
    :param outdir: output directory for mcz-grids
    :return: None
    """
    os.makedirs(outdir, exist_ok=True)
    if isinstance(parameter_table, str):
        parameter_table = pd.read_csv(parameter_table)

    mcz_obs = load_observation(mcz_obs_path).__dict__()

    param_names = parameter_table.columns.tolist()
    n = len(parameter_table)
    n_proc = _get_n_workers()
    if n < n_proc:
        n_proc = n

    kwargs = dict(
        mcz_obs=mcz_obs,
        compas_h5_path=compas_h5_path,
        save_plots=save_images,
        outdir=outdir,
        n_bootstraps=n_bootstraps,
        fname="",
    )
    logger.info(
        f"Generating mcz-grids (with {n_proc} threads for {n} samples of {param_names}) [{kwargs}]"
    )
    # setting up args for process_map
    _lnl_func = partial(McZGrid.lnl, **kwargs)
    iterables = parameter_table.to_dict("records")
    process_map(
        _lnl_func,
        iterables,
        max_workers=n_proc,
        chunksize=n_proc,
    )
    combine_lnl_data(outdir=outdir)


def combine_lnl_data(
    outdir: str = "out_mcz_grids",
    fname: str = "",
) -> None:
    """
    Combine the likelihood data from the generated mcz-grids into a single file
    :param mcz_obs: Observation object containing the mcz data
    :param parameter_table: path to pandas dataframe containing cosmic integration parameters
    :param n_bootstraps: number of bootstraps to generate for each parameter set
    :param outdir: output directory for mcz-grids
    :return: None
    """
    files = glob(f"{outdir}/*_lnl.csv")
    if fname == "":
        fname = f"{outdir}/combined_lnl_data.csv"
    logger.info(f"Compiling {len(files)} LnL values to {outdir}/{fname}")
    df = pd.concat([pd.read_csv(f) for f in files])
    df.to_csv(fname, index=False)


def _get_n_workers():
    """Get the number of workers for parallel processing"""
    total_cpus_available = cpu_count()
    num_workers = 4
    if total_cpus_available > 64:
        num_workers = 16
    elif total_cpus_available > 32:
        num_workers = 8
    elif total_cpus_available < 16:
        num_workers = 4
    logger.warning(
        f"Using {num_workers}/{total_cpus_available} workers for parallel processing"
        "[Total number of CPUs not used to avoid memory issues]"
    )
    # TODO: check -- are we still having memory issues?
    return num_workers



import os
from typing import Dict

import matplotlib.pyplot as plt
import numpy as np


class Observation:
    def __init__(
        self,
        weights: np.ndarray,
        mc_bins: np.array,
        z_bins: np.array,
        duration: float,  # in years
        label: str = "",
        cosmological_parameters: Dict[str, float] = None,
    ):
        self.weights = weights
        self.mc_bins = mc_bins
        self.z_bins = z_bins
        self.duration = duration
        self.n_events, self.n_mc_bins, self.n_z_bins = weights.shape
        assert len(mc_bins) == self.n_mc_bins
        assert len(z_bins) == self.n_z_bins
        self.label = label
        self.cosmological_parameters = cosmological_parameters

    def __repr__(self):
        return f"Obs({self.weights_str(self.weights)}, {self.duration} years)"

    def __str__(self):
        return self.__repr__()

    def __dict__(self):
        d = {
            "weights": self.weights,
            "mc_bins": self.mc_bins,
            "z_bins": self.z_bins,
            "duration": self.duration,
        }
        if self.cosmological_parameters:
            params = self.cosmological_parameters
            params = np.array(
                [(k, v) for k, v in params.items()],
                dtype=[("key", "U10"), ("value", "f8")],
            )
            d["cosmological_parameters"] = params
        d["label"] = self.label
        return d

    def save(self, fname: str = "", outdir: str = ""):
        if outdir:
            fname = os.path.join(outdir, f"{self.label}.npz")
        np.savez(fname, **self.__dict__())

    @classmethod
    def from_dict(cls, data: Dict):
        if "cosmological_parameters" in data:
            params = data["cosmological_parameters"]
            if isinstance(params, np.ndarray):
                params = {k: v for k, v in params}
            data["cosmological_parameters"] = params
        if isinstance(data["label"], np.ndarray):
            data["label"] = data["label"].item()
        if isinstance(data["duration"], np.ndarray):
            data["duration"] = data["duration"].item()
        return cls(**data)

    @classmethod
    def load(cls, fname: str):
        data = np.load(fname)
        loaded_data = {k: data[k] for k in data.files}
        if "label" not in loaded_data:
            loaded_data["label"] = os.path.basename(fname).split(".")[0]
        return cls.from_dict(loaded_data)

    def plot(self, fname=None, ax=None, title=None) -> plt.Figure:
        if ax is None:
            fig, ax = plt.subplots(figsize=(5, 5))
        fig = ax.get_figure()

        Z, MC = np.meshgrid(self.z_bins, self.mc_bins)
        cbar = ax.pcolor(Z, MC, self.weights.sum(axis=0), cmap="inferno")
        ax.set_xlabel(r"$z$")
        ax.set_ylabel(r"$\mathcal{M}_{\rm src}\, [M_{\odot}]$")
        cbar = fig.colorbar(cbar, ax=ax)
        cbar.set_label(r"$\sum_{\rm events} w[i, z, \mathcal{M}]$")
        if title is None:
            title = str(self)
        ax.set_title(title)
        if fname is not None:
            fig.savefig(fname)
        return fig

    @staticmethod
    def weights_str(weights):
        ns, mcs, zs = weights.shape
        return f"n={ns}, bins=[{mcs}, {zs}]"



import os

import matplotlib.pyplot as plt
import numpy as np

from lnl_computer.cosmic_integration.mcz_grid import McZGrid
from lnl_computer.observation.lvk_observation import LVKObservation


def test_lvk_observation(monkeypatched_mcz_grid: McZGrid, tmp_path: str):
    obs = LVKObservation.from_ogc4_data()
    plt_dir = tmp_path + "/obs"
    os.makedirs(plt_dir, exist_ok=True)
    plot_lvk_observation(obs, monkeypatched_mcz_grid, plt_dir, 0)


def plot_lvk_observation(
    obs: LVKObservation, model: McZGrid, plt_dir: str, idx=0
):
    fig, axes = plt.subplots(1, 3, figsize=(15, 5))
    z_bins, mc_bins = model.redshift_bins, model.chirp_mass_bins
    Z, MC = np.meshgrid(z_bins, mc_bins)

    w = obs.weights[idx]
    prob = model.rate_matrix * w

    axes[0].pcolor(Z, MC, w)
    axes[1].pcolor(
        Z,
        MC,
        model.rate_matrix,
    )
    axes[2].pcolor(
        Z,
        MC,
        prob,
    )
    axes[0].set_title("Obs")
    axes[1].set_title("Model")
    axes[2].set_title("Obs*Model")
    for i, m in enumerate([w, model.rate_matrix, prob]):
        labl = f"Sum: {np.sum(m):.2e}"
        axes[i].text(
            0.05,
            0.95,
            labl,
            ha="left",
            va="top",
            color="white",
            transform=axes[i].transAxes,
        )
    plt.savefig(plt_dir + f"/lvk_observation_{idx}.png")



import json
import os
from typing import Dict, List, Union

import click
import pandas as pd
from compas_python_utils.cosmic_integration.binned_cosmic_integrator.bbh_population import (
    generate_mock_bbh_population_file,
)

from .main import (
    batch_lnl_generation,
    combine_lnl_data,
    make_mock_obs,
    make_sf_table,
)


@click.command("make_sf_table")
@click.option(
    "--parameters",
    "-p",
    type=str,
    multiple=True,
    default=["aSF", "dSF", "mu_z", "sigma_0"],
    help="List of parameters to generate mcz-grids for",
)
@click.option(
    "--n",
    "-n",
    type=int,
    default=50,
    help="Number of samples to generate",
)
@click.option(
    "--grid_parameterspace",
    "-g",
    type=bool,
    is_flag=True,
    help="Whether to grid the parameter space",
)
@click.option(
    "--fname",
    "-f",
    type=str,
    default="parameter_table.csv",
    help="Output filename (must be a .csv)",
)
def cli_make_sf_table(
    parameters,
    n,
    grid_parameterspace,
    fname,
) -> None:
    """Parses the table of parameters to generate mcz-grids for."""
    make_sf_table(
        parameters=parameters,
        n=n,
        grid_parameterspace=grid_parameterspace,
        fname=fname,
    )


DEFAULT_SF_PARAMETERS = dict(
    aSF=0.01,
    bSF=2.77,
    cSF=2.9,
    dSF=4.7,
    mu_z=-0.23,
    sigma_0=0.39,
)


@click.command(name="make_mock_obs")
@click.argument("compas_h5_path", type=str)
@click.option(
    "--sf_sample",
    type=str,
    show_default=True,
    default="aSF:0.01 dSF:4.7 mu_z:-0.23 sigma_0:0.39",
    help="Star formation parameters",
)
@click.option(
    "--duration",
    type=float,
    default=1,
    help="Duration of the observation",
    show_default=True,
)
@click.option(
    "--fname",
    type=str,
    default="observation.npz",
    help="Output filename (must be a .npz)",
)
@click.option(
    "--save_plots",
    type=bool,
    is_flag=True,
    default=False,
    help="Save diagnostic plots",
    show_default=True,
)
def cli_make_mock_obs(
    compas_h5_path: str,
    sf_sample: Union[Dict, str],
    duration: float,
    fname: str,
    save_plots: bool,
) -> "MockObservation":
    """Generate a set of 'mock' observations for the sf-sample and compas output file (COMPAS_H5_PATH)."""
    return make_mock_obs(
        compas_h5_path=compas_h5_path,
        duration=duration,
        sf_sample=sf_sample,
        fname=fname,
        save_plots=save_plots,
    )


@click.command(name="batch_lnl_generation")
@click.argument("mcz_obs_path", type=click.Path(exists=True))
@click.argument(
    "compas_h5_path",
    type=str,
)
@click.argument(
    "parameter_table", type=click.Path(exists=True, dir_okay=False)
)
@click.option(
    "--n_bootstraps",
    default=100,
    help="Number of bootstraps to generate for each parameter set (used for error estimation)",
    type=int,
    show_default=True,
)
@click.option(
    "--plots/--no_plots",
    default=True,
    help="Save diagnostic plots for each parameter set",
    is_flag=True,
    show_default=True,
)
@click.option(
    "--outdir",
    default="out_mcz_grids",
    help="Outdir for mcz-grids",
    type=str,
    show_default=True,
)
def cli_batch_lnl_generation(
    mcz_obs_path: str,
    compas_h5_path: str,
    parameter_table: Union[pd.DataFrame, str],
    n_bootstraps: int = 100,
    plots: bool = True,
    outdir: str = "out_mcz_grids",
) -> None:
    """
    Given observations (MCZ_OBS, npz-file), COMPAS output (COMPAS_H5_PATH, h5 file),
    and a table of SF parameters (PARAMETER_TABLE, csv file), generate McZ grids and compute likelihoods.

    The likelihoods are saved to OUTDIR/*_lnl.csv
    """
    batch_lnl_generation(
        mcz_obs_path=mcz_obs_path,
        compas_h5_path=compas_h5_path,
        parameter_table=parameter_table,
        n_bootstraps=n_bootstraps,
        save_images=plots,
        outdir=outdir,
    )


@click.command("combine_lnl_data")
@click.argument("outdir", default="out_mcz_grids", type=str)
@click.option(
    "--fname", default="", type=str, help="Output filename (must be a .csv)"
)
def cli_combine_lnl_data(
    outdir: str = "out_mcz_grids",
    fname: str = "",
) -> None:
    """
    Combine the likelihood data in 'OUTDIR/*_lnl.csv' -> FNAME

    OUTDIR: Output directory with likelihood files (csvs)
    """
    combine_lnl_data(outdir=outdir, fname=fname)


@click.command("mock_compas_output")
@click.argument("fname", type=str, default="mock_compas_output.h5")
def cli_make_mock_compas_output(fname: str):
    """Generate a mock COMPAS output file at FNAME"""
    os.makedirs(os.path.dirname(fname), exist_ok=True)
    generate_mock_bbh_population_file(filename=fname)
    click.echo(f"Mock COMPAS output saved to {fname}")



import os

import numpy as np
import pytest

from lnl_computer.cosmic_integration.mcz_grid import McZGrid
from lnl_computer.mock_data import MockData
from lnl_computer.observation import Observation
from lnl_computer.observation.mock_observation import MockObservation

PLOT = False


def test_mcz_grid_gen_n_save(mock_data: MockData, tmp_path):
    # GENERATE AND SAVE GRID
    grd = McZGrid.from_compas_output(
        mock_data.compas_filename,
        cosmological_parameters=dict(aSF=0.01, dSF=4.70, mu_z=-0.23),
        n_bootstrapped_matrices=2,
        chirp_mass_bins=np.linspace(3, 40, 50),
        redshift_bins=np.linspace(0, 0.6, 100),
    )
    tmp_fn = os.path.join(tmp_path, "mcz_grid.h5")
    grd.save(fname=tmp_fn)


def test_mcz_grid_generation_skip(mock_data: MockData, caplog, tmp_path):
    # TRY TO RE-GENERATE (but will be skipped)
    McZGrid.generate_n_save(
        mock_data.compas_filename,
        dict(aSF=0.01, dSF=4.70, mu_z=-0.23, sigma_0=0.0),
        n_bootstraps=2,
        outdir=tmp_path,
        fname=mock_data.mcz_grid_filename,
    )
    assert f"Skipping {mock_data.mcz_grid_filename} generation" in caplog.text


def test_load_mcz_grid_n_plot(mock_data: MockData, tmp_path):
    new_uni = McZGrid.from_h5(mock_data.mcz_grid_filename)
    new_uni.plot().savefig(os.path.join(tmp_path, "mcz_grid.png"))


def test_obs_gen_n_save(mock_data: MockData, tmp_path):
    obs = MockObservation.from_mcz_grid(mock_data.mcz_grid, duration=1)
    obs.save(os.path.join(tmp_path, "mock_obs.npz"))
    obs = Observation.load(os.path.join(tmp_path, "mock_obs.npz"))
    obs.plot(fname=os.path.join(tmp_path, "mock_obs.png"))



from tempfile import TemporaryDirectory
from typing import Dict

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from compas_python_utils.cosmic_integration.binned_cosmic_integrator.bbh_population import (
    generate_mock_bbh_population_file,
)

from ..cosmic_integration.mcz_grid import McZGrid
from ..cosmic_integration.star_formation_paramters import DEFAULT_DICT
from ..logger import logger
from .observation import Observation


class MockObservation(Observation):
    @classmethod
    def from_mcz_grid(
        cls, mcz_grid: McZGrid, duration: int, n_obs: float = None
    ):
        w = generate_mock_obs_weights(mcz_grid, duration, n_obs)
        return cls(
            w,
            mcz_grid.chirp_mass_bins,
            mcz_grid.redshift_bins,
            duration,
            label=f"MockObs({cls.weights_str(w)})",
            cosmological_parameters=mcz_grid.cosmological_parameters,
        )

    @classmethod
    def from_compas_h5(
        cls,
        compas_h5_fname: str,
        duration: float,
        n_obs: int = None,
        cosmological_parameters=DEFAULT_DICT,
        save_plots: bool = False,
    ):
        mcz_grid = McZGrid.from_compas_output(
            compas_h5_fname,
            cosmological_parameters,
            save_plots=save_plots,
        )
        mcz_grid.bin_data()
        return cls.from_mcz_grid(mcz_grid, duration, n_obs)

    def __repr__(self):
        return "Mock" + super().__repr__()


def generate_mock_obs_weights(
    mcz_grid: McZGrid, duration: float, n_obs: int = None
):
    mcz_samples = _sample_events_from_mcz_grid(
        mcz_grid, duration=duration, n_obs=n_obs
    )
    w = np.zeros((len(mcz_samples), *mcz_grid.rate_matrix.shape))
    for i, (mc, z) in enumerate(mcz_samples):
        mc_bin, z_bin = mcz_grid.get_matrix_bin_idx(mc, z)
        w[i, mc_bin, z_bin] += 1
    return w


def _sample_events_from_mcz_grid(
    mcz_grid: McZGrid,
    duration: float,
    n_obs: float = None,
) -> np.ndarray:
    """
    Sample Mc-Z pairs from the mcz_grid.

    Sample using the detection rate as weights (i.e. sample more from higher detection rate regions).
    #TODO: implement sample_using_emcee (sample from the detection rate distribution using poisson distributions)
    #TODO: draw from the detection rate distribution using poisson distributions

    :param mcz_grid: The mcz_grid to sample from
    :param duration: The duration of the observation (in years), to convert rate->number of events
    :param n_obs: The number of observations to sample (could be a float, _will_not_ be rounded to an int)

    :return: np.ndarray of shape (n_obs, 2) where each row is (mc, z)
    """
    n_obs = mcz_grid.n_detections(duration) if n_obs is None else n_obs
    logger.info(
        f"Sampling {n_obs:.1f} events ({duration}yrs) from mcz_grid[{mcz_grid}]"
    )
    df = _mcz_to_df(mcz_grid)
    if np.sum(df.rate) > 0:
        n_events = df.sample(
            weights=df.rate, n=int(n_obs), random_state=0, replace=True
        )
    else:
        n_events = df.sample(n=int(n_obs), random_state=0)

    return n_events[["mc", "z"]].values


def _mcz_to_df(mcz_grid) -> pd.DataFrame:
    """The mcz_grid as a pandas dataframe with columns (mc, z, rate), sorted by rate (high to low)"""

    z, mc = mcz_grid.redshift_bins, mcz_grid.chirp_mass_bins
    rate = mcz_grid.rate_matrix.ravel()
    zz, mcc = np.meshgrid(z, mc)
    df = pd.DataFrame({"z": zz.ravel(), "mc": mcc.ravel(), "rate": rate})
    df = df.sort_values("rate", ascending=False)

    # drop nans and log the number of rows dropped
    n_nans = df.isna().any(axis=1).sum()
    if n_nans > 0:
        logger.warning(f"Dropping {n_nans}/{len(df)} rows with nan values")
        df = df.dropna()

    # check no nan in dataframe
    if df.isna().any().any():
        logger.error("Nan values in dataframe")

    return df



import os
import shutil

import pandas as pd
import pytest
from click.testing import CliRunner

from lnl_computer.cli.cli import (
    cli_batch_lnl_generation,
    cli_combine_lnl_data,
    cli_make_mock_obs,
    cli_make_sf_table,
    make_mock_obs,
    make_sf_table,
)
from lnl_computer.mock_data import MockData
from lnl_computer.observation import Observation
from lnl_computer.observation.mock_observation import MockObservation


def test_cli_make_sf_table(tmp_path):
    fname = f"{tmp_path}/parameter_table.csv"
    runner = CliRunner()
    n = 2**5
    params = ["aSF", "dSF"]
    out = runner.invoke(
        cli_make_sf_table,
        ["-p", params[0], "-p", params[1], "-n", n, "-f", fname],
    )
    print(out.stdout)
    assert out.exit_code == 0
    df = pd.read_csv(fname)
    assert len(df) == n
    assert set(df.columns) == set(params)


def test_cli_make_mock_obs(mock_data: MockData, tmp_path):
    fname = f"{tmp_path}/mock_obs.npz"
    runner = CliRunner()
    asf = 0.01
    out = runner.invoke(
        cli_make_mock_obs,
        [
            mock_data.compas_filename,
            "--sf_sample",
            f"aSF:{asf}",
            "--fname",
            fname,
        ],
    )
    print(out.stdout)
    assert out.exit_code == 0
    obs = Observation.load(fname)
    assert obs.cosmological_parameters["aSF"] == asf


def test_cli_batch_lnl_generation(mock_data, tmp_path):
    # STEP 1: make a parameter table
    tmp_path = f"{tmp_path}/lnl_generation"
    shutil.rmtree(tmp_path, ignore_errors=True)
    os.makedirs(tmp_path, exist_ok=True)
    sf_fname = f"{tmp_path}/parameter_table.csv"
    make_sf_table(parameters=["aSF", "dSF"], n=2, fname=sf_fname)
    sf_parm = pd.read_csv(sf_fname).to_dict("records")[0]

    # STEP 2: generate mock observations
    mock_fname = f"{tmp_path}/mock_obs.npz"
    make_mock_obs(
        mock_data.compas_filename, sf_parm, fname=mock_fname, duration=1
    )

    # STEP 3: generate lnl data
    runner = CliRunner()
    out = runner.invoke(
        cli_batch_lnl_generation,
        [
            mock_fname,
            mock_data.compas_filename,
            sf_fname,
            "--n_bootstraps",
            1,
            "--no_plots",
            "--outdir",
            tmp_path,
        ],
    )
    assert out.exit_code == 0, f"Error in batch_lnl_generation: {out.stdout}"

    # STEP 4: combine lnl data
    comb_fname = f"{tmp_path}/combined_lnl_data.csv"
    out = runner.invoke(
        cli_combine_lnl_data, [tmp_path, "--fname", comb_fname]
    )
    data = pd.read_csv(comb_fname)
    assert len(data) == 2
    assert out.exit_code == 0, f"Error in combine_lnl_data: {out.stdout}"



from lnl_computer.observation import Observation, load_observation


def test_load_obs(tmpdir, mock_data):

    obs = load_observation("LVK")
    assert obs is not None
    assert isinstance(obs, Observation)

    obs = load_observation(mock_data.observations_filename)
    assert obs is not None
    assert isinstance(obs, Observation)

    data_dict = obs.__dict__()

    obs1 = Observation.from_dict(data_dict)
    obs2 = Observation.from_dict(data_dict)
    assert isinstance(obs, Observation)
    assert obs1.duration == obs2.duration



import numpy as np
from ogc4_interface.population_mcz import PopulationMcZ

from .observation import Observation


class LVKObservation(Observation):
    @classmethod
    def from_ogc4_data(
        cls,
        pastro_threshold=0.95,
        observing_runs=["O3a", "O3b"],
        filter_valid_mcz=True,
    ) -> "LVKObservation":
        data = PopulationMcZ.load(
            pastro_threshold=pastro_threshold,
            observing_runs=observing_runs,
            filter_valid_mcz=filter_valid_mcz,
        )
        w = data.weights
        # change from (n_events, z_bins, mc_bins) to (n_events, mc_bins, z_bins)
        w = np.moveaxis(w, (0, 1, 2), (0, 2, 1))
        ne, nmc, nz = w.shape
        return cls(
            w,
            mc_bins=data.mc_bins,
            z_bins=data.z_bins,
            duration=data.duration,
            label=f"LVKObs({cls.weights_str(w)})]",
        )

    def __repr__(self):
        return "LVK" + super().__repr__()



import os
import shutil
from typing import Dict, Tuple, Union

import numpy as np
import pandas as pd
from compas_python_utils.cosmic_integration.binned_cosmic_integrator.detection_matrix import (
    DetectionMatrix,
)

from ..likelihood import ln_likelihood
from ..logger import logger
from ..observation import Observation, load_observation
from .star_formation_paramters import DEFAULT_SF_PARAMETERS


class McZGrid(DetectionMatrix):
    """
    Represents a detection matrix in (m_c, z) space.

    This class is a wrapper around the DetectionMatrix class, with some extra functionality.

    :param compas_path: The path to the COMPAS hdf5 file
    :param cosmological_parameters: The cosmological parameters to use (dict)
    :param rate_matrix: The detection rate matrix
    :param chirp_mass_bins: The chirp mass bins
    :param redshift_bins: The redshift bins
    :param n_systems: The number of systems in the COMPAS file
    :param n_bbh: The number of BBH systems in the COMPAS file
    :param outdir: The output directory to save the detection matrix to
    :param bootstrapped_rate_matrices: The bootstrapped rate matrices


    Important methods:
    - sample_observations
    - prob_of_mcz

    """

    @classmethod
    def from_dict(cls, data: Dict) -> "McZGrid":
        """Create a mcz_grid object from a dictionary"""
        obj_types = [
            "compas_path",
            "cosmological_parameters",
            "n_systems",
            "n_bbh",
        ]
        for key in data:
            if key in obj_types:
                data[key] = data[key].item()
        det_matrix = cls(
            compas_path=data["compas_path"],
            cosmological_parameters=data["cosmological_parameters"],
            rate_matrix=data["rate_matrix"],
            chirp_mass_bins=data["chirp_mass_bins"],
            redshift_bins=data["redshift_bins"],
            n_systems=data["n_systems"],
            n_bbh=data["n_bbh"],
            outdir=data.get("outdir", "."),
            bootstrapped_rate_matrices=data.get(
                "bootstrapped_rate_matrices", None
            ),
        )
        logger.debug(f"Loaded cached det_matrix with: {det_matrix.param_str}")
        return det_matrix

    def save(self, fname="") -> None:
        """Save the mcz_grid object to a npz file, return the filename"""
        super().save()
        if fname != "":
            shutil.move(self.default_fname, fname)

    def get_matrix_bin_idx(self, mc: float, z: float) -> Tuple[int, int]:
        mc_bin = np.argmin(np.abs(self.chirp_mass_bins - mc))
        z_bin = np.argmin(np.abs(self.redshift_bins - z))
        return mc_bin, z_bin

    def prob_of_mcz(self, mc: float, z: float, duration: float = 1.0) -> float:
        mc_bin, z_bin = self.get_matrix_bin_idx(mc, z)
        if self.n_detections(duration) == 0:
            return 0
        return self.rate_matrix[mc_bin, z_bin] / self.n_detections(duration)

    def get_prob_grid(self, duration: float = 1.0) -> np.ndarray:
        if self.n_detections(duration) == 0:
            return np.zeros_like(self.rate_matrix)
        return self.rate_matrix / self.n_detections(duration)

    def get_bootstrapped_grid(self, i: int) -> "McZGrid":
        """Creates a new Uni using the ith bootstrapped rate matrix"""
        assert (
            i < self.n_bootstraps
        ), f"i={i} is larger than the number of bootstraps {self.n_bootstraps}"
        return McZGrid(
            compas_path=self.compas_path,
            cosmological_parameters=self.cosmological_parameters,
            rate_matrix=self.bootstrapped_rate_matrices[i],
            chirp_mass_bins=self.chirp_mass_bins,
            redshift_bins=self.redshift_bins,
            n_systems=self.n_systems,
            n_bbh=self.n_bbh,
            outdir=self.outdir,
        )

    def n_detections(self, duration: float = 1.0) -> float:
        """Calculate the number of detections in a given duration (in years)"""
        return np.nansum(self.rate_matrix) * duration

    def get_lnl(
        self,
        mcz_obs: Observation,
    ) -> Tuple[float, float]:
        """Get Lnl+/-unc from the mcz_obs"""
        lnl = ln_likelihood(
            obs=mcz_obs,
            model=self,
        )
        bootstrapped_lnls = []
        for i in range(self.n_bootstraps):
            bootstrap_mcz_grid = self.get_bootstrapped_grid(i)
            bootstrapped_lnls.append(
                ln_likelihood(
                    obs=mcz_obs,
                    model=bootstrap_mcz_grid,
                )
            )
        return lnl, np.std(np.array(bootstrapped_lnls))

    def __dict__(self) -> Dict:
        return self.to_dict()

    def __repr__(self) -> str:
        return f"<mcz_grid: [{self.n_systems} systems], {self.param_str}, [{self.compas_path}]>"

    @property
    def param_str(self):
        return "_".join(
            [f"{k}_{v:.10f}" for k, v in self.cosmological_parameters.items()]
        )

    @property
    def label(self) -> str:
        return McZGrid.get_label(
            self.compas_path, self.cosmological_parameters
        )

    @property
    def default_fname(self) -> str:
        return McZGrid.get_default_fname(
            self.outdir, self.compas_path, self.cosmological_parameters
        )

    @staticmethod
    def get_label(
        compas_path: str, cosmological_parameters: Dict[str, float]
    ) -> str:
        compas_fname = os.path.basename(compas_path).split(".")[0]
        p_str = McZGrid.param_dict_to_str(cosmological_parameters)
        return f"mczgrid_{compas_fname}_{p_str}"

    @staticmethod
    def get_default_fname(
        outdir: str,
        compas_path: str,
        cosmological_parameters: Dict[str, float],
    ):
        l = McZGrid.get_label(compas_path, cosmological_parameters)
        return f"{outdir}/{l}.h5"

    @staticmethod
    def param_dict_to_str(param_dict: Dict[str, float]) -> str:
        return "_".join([f"{k}_{v:.4f}" for k, v in param_dict.items()])

    @property
    def n_bootstraps(self) -> int:
        if self.bootstrapped_rate_matrices is None:
            return 0
        return len(self.bootstrapped_rate_matrices)

    @classmethod
    def generate_n_save(
        cls,
        compas_h5_path: str,
        sf_sample: Dict = None,
        save_plots: bool = False,
        outdir=None,
        fname="",
        n_bootstraps=0,
        clean=False,
        **kwargs,
    ) -> "McZGrid":
        """Generate a detection matrix for a given set of star formation parameters
        :param compas_h5_path:
        :param sf_sample: Dict of star formation parameters
        :param save_plots: Bool to save plots
        :param outdir: outdir for plots + mcz_grid
        :param fname: mcgrid-fname (if empty, will not save)
        :param n_bootstraps: N
        :return:
        """

        if sf_sample is None:
            sf_sample = DEFAULT_SF_PARAMETERS
            logger.warning(
                f"sf_sample not provided, using default: {sf_sample}"
            )

        if not clean:
            fnames = [
                fname,
                cls.get_default_fname(outdir, compas_h5_path, sf_sample),
            ]
            for f in fnames:
                if os.path.isfile(f):
                    logger.warning(
                        f"Skipping {f} generation as it already exists"
                    )
                    return cls.from_h5(f)
        if fname != "" and not fname.endswith(".h5"):
            logger.error(f"fname must end with .h5, got {fname}")

        if sf_sample is None:
            sf_sample = DEFAULT_SF_PARAMETERS

        params = dict(
            aSF=sf_sample.get("aSF", DEFAULT_SF_PARAMETERS["aSF"]),
            dSF=sf_sample.get("dSF", DEFAULT_SF_PARAMETERS["dSF"]),
            mu_z=sf_sample.get("mu_z", DEFAULT_SF_PARAMETERS["mu_z"]),
            sigma_0=sf_sample.get("sigma_0", DEFAULT_SF_PARAMETERS["sigma_0"]),
        )
        logger.info(f"Generating McZ grid: {params}")

        mcz_grid = cls.from_compas_output(
            compas_path=compas_h5_path,
            cosmological_parameters=params,
            max_detectable_redshift=0.6,
            redshift_bins=np.linspace(0, 0.6, 100),
            chirp_mass_bins=np.linspace(3, 40, 50),
            outdir=outdir,
            save_plots=save_plots,
            n_bootstrapped_matrices=n_bootstraps,
        )
        if isinstance(fname, str):
            mcz_grid.save(fname=fname)
        return mcz_grid

    @classmethod
    def lnl(
        cls,
        sf_sample: Dict,
        mcz_obs: Union[Observation, str, Dict],
        compas_h5_path: str,
        **kwargs,
    ) -> Tuple[float, float]:
        """Return the LnL(sf_sample|mcz_obs)+/-unc

        Also saves the Lnl+/-unc and params to a csv file

        :param mcz_obs: The observed mcz values
        :param args: Arguments to pass to generate_n_save
        :return: The LnL value
        """
        if "outdir" in kwargs:
            os.makedirs(kwargs["outdir"], exist_ok=True)

        if isinstance(mcz_obs, str):
            mcz_obs = load_observation(mcz_obs)
        elif isinstance(mcz_obs, dict):
            mcz_obs = Observation.from_dict(mcz_obs)

        model = cls.generate_n_save(
            compas_h5_path=compas_h5_path,
            sf_sample=sf_sample,
            **kwargs,
        )

        lnl, unc = model.get_lnl(mcz_obs=mcz_obs)
        _save_lnl_dict_to_csv(
            lnl,
            unc,
            model=model,
            fname=kwargs.get("fname", ""),
        )
        return lnl, unc

    @property
    def n_mc_bins(self) -> int:
        return len(self.chirp_mass_bins)

    @property
    def n_z_bins(self) -> int:
        return len(self.redshift_bins)


def _save_lnl_dict_to_csv(lnl, unc, model: McZGrid, fname: str) -> None:
    """Save the lnl and unc to a csv file"""
    data = dict(
        lnl=lnl,
        unc=unc,
        **model.cosmological_parameters,
    )
    # save lnl data to csv
    lnl_fname = f"{model.outdir}/{model.label}_lnl.csv"
    if fname != "":
        lnl_fname = fname.replace(".h5", "_lnl.csv")

    df = pd.DataFrame(data, index=[0])
    df.to_csv(lnl_fname, index=False)
